#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "define.h"

#define _OUTPUT_MATRIX		0
#define _OUTPUT_CURVE		0
#define _OUTPUT_RADIANCE	1

#define R_INDEX				2
#define G_INDEX				1
#define B_INDEX				0

double weight(unsigned char z)
{
	if(z <= ZMID)
		return (double)((z - ZMIN)/2 + 128.0);
	else
		return (double)((ZMAX - z)/2 + 128.0);
}

void hdr_generate(bmp **b, config *conf, hdr *h, double *g[3])
{
	int i;
	int j;

	h->xresolution = b[0]->info.width;
	h->yresolution = b[0]->info.height;
	int size = h->xresolution * h->yresolution;
	
	int in_idx = 0;

	double lnR, lnG, lnB;
	double wR_temp, wG_temp, wB_temp;
	double wR, wG, wB;

	h->s = new irradiance[size];
	
	for(i = 0; i < size; i++)
	{
		lnR = lnG = lnB = 0.0;
		wR  = wG  = wB  = 0.0;
		for(j = 0; j < conf->frame_number; j++)
		{
			wR_temp = weight(b[j]->data[in_idx+R_INDEX]);
			wG_temp = weight(b[j]->data[in_idx+G_INDEX]);
			wB_temp = weight(b[j]->data[in_idx+B_INDEX]);						

			lnR += wR_temp * ( g[R_INDEX][ b[j]->data[in_idx+R_INDEX]] - conf->lnt[j]);
			lnG += wG_temp * ( g[G_INDEX][ b[j]->data[in_idx+G_INDEX]] - conf->lnt[j]);
			lnB += wB_temp * ( g[B_INDEX][ b[j]->data[in_idx+B_INDEX]] - conf->lnt[j]);

			wR += wR_temp;
			wG += wG_temp;
			wB += wB_temp;
		}
		h->s[i].r = exp(lnR/wR);
		h->s[i].g = exp(lnG/wG);
		h->s[i].b = exp(lnB/wB);
		in_idx = in_idx + 3;
	}

#if(_OUTPUT_RADIANCE)
	// Output G only;
	FILE *output = fopen("RadianceMap.txt", "w");
	for(i = b[0]->info.height-1 ; i >= 0 ; i--)
	{
		for(j = 0; j < b[0]->info.width; j++)
		{
			fprintf(output, "%4f ", (float) h->s[i*b[0]->info.width+j].g);
		}
		fprintf(output, "\n");
	}
	fclose(output);
#endif
}


int main(int argc, char *argv[])
{
	int i, j, k, c;								// Loop counters

	double *g[3];						// Inverse respone curves of R, G, B
	bmp    **image;						// Images
	bmp		*info;
	
	if(argc < 3)
	{
		printf("Please specify a config and an output files\n");
		exit(1);
	}

	// Load config file
	config *conf = load_config(argv[1]);

	// Load images
	image = new bmp *[conf->frame_number];
	for(i = 0; i < conf->frame_number; i++)
	{
		image[i] = new bmp;
		printf("Read frame %d\n", i);				
		bmp_read(conf->image_name[i], image[i]);
	}
	info = image[0];

	// Variables used in [Debevec] Appendix A
	int num_pixel = 4*(ZMAX-ZMIN)/conf->frame_number;
	//int num_pixel = 100;
	int num_image = conf->frame_number;
	int n = 256;
	int size;
	unsigned char **Z;
	double **A;
	double *b;
	double lamda = LAMDA;
	double wij;

	size = num_pixel*num_image + n - 2;

	// Variables used in [NP] Chapter 2
	int		M = size;
	int		N = n + num_pixel - 1;
	double *x;
	double *D = new double [N+1];
	double **V;
	x = new double[N+1];
	V = new double*[N+1];
	for(i = 0; i <= N; i++)
	{	
		V[i] = new double [N+1];
	}	

	A = new double*[size+1];
	b = new double[size+1];
	for(i = 0; i <= size; i++)
	{
		A[i] = new double[n + num_pixel + 1];
	}

	// Select pixels !!!! Important
	int *idx = new int [num_pixel];
	int randx, randy;
	Z = new unsigned char*[num_pixel+1];

	for(i = 0; i < num_pixel; i++)
	{
		Z[i+1] = new unsigned char[num_image+1];
		randx = (int)(0.2*image[0]->info.width + rand() * 0.8 * image[0]->info.width)/RAND_MAX;
		randy = (int)(0.2*image[0]->info.height + rand()* 0.8 * image[0]->info.height)/RAND_MAX;

		idx[i] = (randy*image[0]->info.width + randx)*3;
	}
	// Data fitting equation
	for(c = 0; c < 3; c++)
	{
		printf("Generate data fitting equation of channel %d\n", c);

		// Set A, b to zero
		for(i = 1; i <= M; i++)
		{			
			for(j = 1; j <= N; j++)
			{				
				A[i][j] = 0.0;
				D[j] = 0.0;
				if(i <= N)
				{
					V[i][j] = 0.0;
					x[i] = 0.0;
				}
			}			
			b[i] = 0.0;
		}
		k = 1;
		for(i = 1; i <= num_pixel; i++)
		{
			for(j = 1; j <= num_image; j++)
			{
				Z[i][j] = image[j-1]->data[idx[i-1]+c];// Z[i][j] ith pixel in jth image				
				wij = weight(Z[i][j]);

				A[k][Z[i][j]+1] = wij;
				A[k][n+i]		= -wij;
				b[k]			= wij * conf->lnt[j-1];

				k = k + 1;
			}
		}

		// Fix the scaling
		A[k][129] = 1;
		k = k + 1;
		
		// Smoothness equations
		for(i = 1; i < n-2; i++)
		{
			A[k][i]		= lamda * weight(i+1);
			A[k][i+1]	= -2 * lamda * weight(i+1);
			A[k][i+2]	= lamda*weight(i+1);
			k = k + 1;
		}

#if(_OUTPUT_MATRIX)
		FILE *MA;
		FILE *bb;
		switch(c)
		{
			case 0:	MA = fopen("MatrixB.txt", "w"); bb = fopen("bB.txt", "w"); break;
			case 1:	MA = fopen("MatrixG.txt", "w"); bb = fopen("bG.txt", "w"); break;
			case 2: MA = fopen("MatrixR.txt", "w"); bb = fopen("bR.txt", "w"); break;
		}
		for(i = 1; i <= M; i++)
		{
			for(j = 1; j <= N; j++)
			{
				fprintf(MA, " %5f", A[i][j]);
			}
			fprintf(MA, "\n");
			fprintf(bb, " %5f\n", b[i]);
		}
		fclose(MA);
		fclose(bb);

#endif
		// Solve the problem
		printf("Solve the problem of channel %d\n", c);
		svd_build(A, M, N, D, V);
		svd_solve(A, D, V, M, N, b, x);	

		// Output the matrix for Matlab
		g[c] = new double[n];
		for(i = 0; i < n; i++)
		{
			g[c][i] = x[i+1];
		}
	}
	
	// inverse curves, map g[i][128] = 0;
	double shift;
	for(i = 0; i < 3; i++)
	{
		shift = 0;// - g[i][n/2];
		for(j = 0; j < n; j++)
		{
			g[i][j] += shift;
		}
	}

#if(_OUTPUT_CURVE)
	FILE *curve;
	curve = fopen("curve.txt", "w");
	for(i = 0; i < n; i++)
	{
		fprintf(curve, "%3.5f %3.5f %3.5f\n", (float)g[R_INDEX][i], (float)g[G_INDEX][i], (float)g[B_INDEX][i]);
	}
	fclose(curve);
#endif

	printf("Generate the radiance map\n");
	hdr *h = new hdr;
	hdr_generate(image, conf, h, g);

	printf("Write the radiance map\n");
	hdr_write(argv[2], h);

	return 1;
}